#! /usr/bin/env python

"""Utilities for FASTR related activities"""

__version__ = '0.0.0.2'


def compress_fastr(sequence):
    """Eliminates repeats in a sequence and returns a string"""

    compressed_sequence = []
    if type(sequence) == str:
        sequence = sequence.split('-')
    last_read = None
    read_repeats = 0
    for read in enumerate(sequence):
        if read[1] == last_read:
            read_repeats += 1
        else:
            if read_repeats != 0:
                compressed_sequence.append(
                    '{0}x{1}'.format(str(read_repeats + 1), str(last_read)))
                read_repeats = 0
            elif read[0] != 0:
                compressed_sequence.append(str(last_read))
            last_read = read[1]
        if read[0] + 1 == len(sequence):
            if read_repeats != 0:
                compressed_sequence.append(
                    '{0}x{1}'.format(str(read_repeats + 1), str(last_read)))
            else:
                compressed_sequence.append(str(last_read))
            compressed_sequence = '-'.join(compressed_sequence)
    return compressed_sequence


def decompress_fastr(sequence):
    """Returns decompressed FASTR sequence"""

    depth_sequence = sequence.split('-')
    decompressed_sequence = []
    for base in depth_sequence:
        try:
            int(base)
            decompressed_sequence.append(int(base))
        except ValueError:
            sep = base.split('x')
            sep[1] = sep[1].rstrip('\n')
            decompressed_sequence += [int(sep[1]) for i in range(int(sep[0]))]
    decompressed_sequence = '-'.join([str(i) for i in decompressed_sequence])
    return decompressed_sequence


def write_fastr(sequence_ids, sequences, fastr_handle):
    """Writes the list of sequence ids and sequences to a FASTR file"""

    for sequence_id, sequence in zip(sequence_ids, sequences):
        fastr_entry = '+{0}\n{1}\n'.format(sequence_id, sequence)
        fastr_handle.write(fastr_entry)



#! /usr/bin/env python

"""metameta 0.0.0.43 - meta-transcriptome/genome mapping analysis toolkit

Usage:

    metameta <tool> [arguments for tool]

Synopsis:

    metameta is a toolkit with a number of programs useful for analyzing
    meta-transcriptome/genome data that has been mapped to a reference.
    It is contains a number of generally useful bioinformatic scripts.
    Type "metameta --docs" for full documentation.

Options:

    Toolkit (options from metameta, no tool specified/needed):

        --help             This help
        --version          Print toolkit version and exits
        --docs             Show full manual

    General Tool Options (options shared by most or all tools):

        no arguments       Gives tool help and exits
        --help             Gives tool help and exits
        --version          Prints tool version and exits
        --verify           Verify relevant files before executing
        --log_file         Write output to the log file

    Verbosity Settings (all tools):

        none: Fatal errors only
        -v: Fatal errors and important information only
        -vv: Detailed information on everything the program is doing,
             best reserved for debugging purposes
        Note: any numbers of "v"s may be specified but anything greater than
        -vv will be identical to -vv.

Tools:

    compute_gene_abundance:

            Reads a GFF3 file and a FASTR or BAM file to produce a
            tab-delimited file containing genes and their abundance values.

    generate_fastr:

            Generates a FASTR file containing per base read depth data
            of a given FASTA or FASTQ file.
"""

__version__ = '0.0.0.43'

import argparse
import metameta.bin
from pkg_resources import resource_string
import subprocess
import sys


def tool_check(desired_tool):
    """Checks if given tool is valid and prints a list of tools if not"""

    tools = [
        'compute_gene_abundance',
        'generate_fastr',
    ]
    for tool in tools:
        if desired_tool == tool:
            return desired_tool
    print('No such tool: ' + desired_tool)
    print('Available tools:')
    for tool in tools:
        print(tool)
    sys.exit(1)


def main():
    parser = argparse.ArgumentParser(description=__doc__,
                                     formatter_class=argparse.
                                     RawDescriptionHelpFormatter)
    parser.add_argument('tool',
                        type=tool_check,
                        default=None,
                        nargs='?',
                        help='tool to run')
    parser.add_argument('arguments',
                        nargs=argparse.REMAINDER,
                        help='arguments to pass to tool')
    parser.add_argument('--docs',
                        help='prints documentation and exits',
                        action='store_true')
    parser.add_argument('--version',
                        action='version',
                        version=__version__)
    args = parser.parse_args()

    location = metameta.bin.__file__.rstrip('__init__.pyc')
    if args.docs:
        documentation = resource_string('metameta.data', 'Documentation.txt')
        subprocess.call(['less', documentation])
    elif args.tool is None:
        print(__doc__)
    else:
        script = location + args.tool + '.py'
        subprocess.call(['python', script, str(args.arguments)])


if __name__ == '__main__':
    main()
    sys.exit(0)



#! /usr/bin/env python

"""print statement with improved functionality"""

from __future__ import print_function

__version__ = '0.0.0.3'

from datetime import datetime
import sys


def output(message, program_verbosity, message_verbosity, log_file=None,
           time_stamp=True, fatal=False):
    """Writes verbosity dependant message to either STDOUT or a log file"""

    if int(program_verbosity) >= int(message_verbosity):
        if time_stamp:
            message_time = '[{0}] '.format(str(datetime.now()))
            message = message_time + message
        if fatal:
            fatal_message = '\nAbove error is fatal. Exiting program.'
            message += fatal_message
        if log_file is None:
            print(message)
        else:
            with open(log_file, 'a') as out_handle:
                out_handle.write(message + '\n')
        if fatal:
            sys.exit(1)



#! /usr/bin/env python

"""calculates gene abundance data from annotations of an assembly

Usage:

    compute_gene_abundance.py [--bam] [--delimiter] [--fasta] [--fastr]
                              [--normalize] <gff3> <ouput>

Synopsis:

    calculates and normalizes gene abundance of annotated genes in a GFF3
    file using a variety of files to obtain abundance data from:
    BAM files
    FASTA files from IDBA-UD
    FASTR files

Required Arguments:

    gff3             GFF3 formatted annotation file
    output           CSV file containing gene ID and abundance,
                     default delimiter is "\t"

Optional Arguments:

    --bam            BAM file containing alignment data
    --delimiter      the character to delimit the output file by [Default: '\t']
    --fasta          IDBA-UD generated FASTA file used for assembly
    --fastr          FASTR file containing read depth data
    --normalize      Method to calculate coverage values (see below for more
                     information on normalization) [Default: 'read_count']

Supported Normalization Methods:

    arpb             Average Reads Per Base: average depth coverage of each base
    read_count       Uses the number of reads used to assemble contig as
                     coverage [Default]
    rpk              Reads Per Kilobase: number of reads used to assemble
                     contig divided by kilobases in the contig (contig length
                     divided by 1,000)
    rpkt, rpb        rpk times 1,000 (equivalent to reads per base)
    rpkm             rpk times 1,000,000
    rpkb             rpk times 1,000,000,000
"""

from __future__ import print_function

__version__ = '0.0.0.1'

import argparse
from bio_utils.iterators.gff3 import gff3_iter
from bio_utils.iterators.fastr import fastr_iter
from bio_utils.verifiers.binary import binary_verifier
from bio_utils.verifiers.fasta import fasta_verifier
from bio_utils.verifiers.fastr import fastr_verifier
from bio_utils.verifiers.gff3 import gff3_verifier
from metameta.metameta_utils.fastr_utils import decompress_fastr
from metameta.metameta_utils.output import output
import pysam
import re
from screed.fasta import fasta_iter
import statistics
import sys


def compute_gene_abundance_from_bam(bam_file, gff3_file, database,
                                    normalization, out_file):
    with open(out_file, 'a') as out_handle:
        with pysam.AlignmentFile(bam_file, 'rb') as bam_handle:
            with open(gff3_file, 'rU') as gff3_handle:
                for entry in gff3_iter(gff3_handle):
                    db_id = extract_db_id(entry['attributes'], database)
                    if entry['start'] < entry['end']:
                        start = int(entry['end']) - 1
                        end = int(entry['start']) - 1
                    else:
                        start = int(entry['start']) - 1
                        end = int(entry['end']) - 1
                    if normalization == 'arpb':
                        reads_per_base = [pileUp.n for pileUp in bam_handle.
                                          pileup(entry['seq_id'], start, end)]
                        coverage = normalize(normalization,
                                             per_base_depth=reads_per_base)
                    else:
                        read_count = len([read for read in bam_handle.fetch(
                                          entry['seqid'], start, end)])
                        gene_length = end - start
                        coverage = normalize(normalization,
                                             read_count=read_count,
                                             length=gene_length)
                    out_handle.write('{0}\t{1}\n'.format(db_id, coverage))


def compute_gene_abundance_from_fasta(fasta_file, gff3_file, database,
                                      normalization, out_file):
    gff3_dict = gff3_to_dict(gff3_file, database)
    with open(out_file, 'a') as out_handle:
        with open(fasta_file, 'rU') as fasta_handle:
            for entry in fasta_iter(fasta_handle):
                read_count = entry['description'].split('read_count_')[-1]
                if normalization != 'read_count':
                    contig_length = len(entry['sequence'])
                    read_count = normalize(normalization,
                                           length=contig_length,
                                           read_count=read_count)
                coverage = read_count
                for gene in gff3_dict[entry['name']]:
                    db_id = extract_db_id(gene['attributes'], database)
                    out_handle.write('{0}\t{1}\n'.format(db_id, coverage))


def compute_gene_abundance_from_fastr(fastr_file, gff3_file, database,
                                      normalization, out_file):
    if normalization != 'arpb':
        message = 'FASTR files can only be used to calculate coverage as ' \
                  'Average Reads Per Base (arpb). Run this program with the' \
                  ' "--normalzie arpb" flag or use a different file type to ' \
                  'calculate coverage from.'
        output(message, 0, 0, fatal=True)
    gff3_dict = gff3_to_dict(gff3_file, database)
    with open(out_file, 'a') as out_handle:
        with open(fastr_file, 'rb') as fastr_handle:
            for entry in fastr_iter(fastr_handle):
                if entry['name'] in gff3_dict:
                    fastr_sequence = decompress_fastr(entry['sequence'])
                    coverage_sequence = [int(base) for base in
                                         fastr_sequence.split('-')]
                    for gene in gff3_dict[entry['name']]:
                        if gene['start'] < gene['end']:
                            start = int(gene['end']) - 1
                            end = int(gene['start']) - 1
                        else:
                            start = int(gene['start']) - 1
                            end = int(gene['end']) - 1
                        reads_per_base = coverage_sequence[start:end]
                        coverage = normalize(normalization,
                                             per_base_depth=reads_per_base)
                        db_id = extract_db_id(gene['attributes'], database)
                        out_handle.write('{0}\t{1}\n'.format(db_id, coverage))


def extract_db_id(attributes, database):
    find_id = re.compile('{0}:.*?;'.format(database))
    id_match = re.match(find_id, attributes)
    if len(str(id_match)) == 1:
        db_id = str(id_match).lstrip('{0}:'.format(database)).rstrip(';')
        return db_id


def gff3_to_dict(gff3_file, database):
    gff3_dict = {}
    with open(gff3_file, 'rU') as gff3_handle:
        for entry in gff3_iter(gff3_handle):
            db_id = extract_db_id(entry['attributes'], database)
            gff3_dict[entry['seqid']][db_id] = entry
    return gff3_dict


def normalization_method(method):
    acceptable_methods = ['read_count', 'rpk', 'rpkt', 'rpkm', 'rpkb', ' rpb',
                          'arpb']
    if method in acceptable_methods:
        return str(method)
    else:
        message = '{0} is not a supported normalization method. Supported ' \
                  'normalization methods follow:\n{1}'.format(method,
                  '\n'.join(acceptable_methods))
        output(message, 0, 0, fatal=True)


def normalize(normalization, length=0, read_count=0,
              per_base_depth=None):
    normalized_coverage = False
    if normalization == 'read_count':
        return read_count
    elif 'rpk' in normalization:
        normalization_factor = {
            'rpk': 1,
            'rpkt': 1000,
            'rpkm': 1000000,
            'rpkb': 1000000000
        }
        normalized_coverage = (read_count / (length / 1000)) * \
                               normalization_factor[normalization]
    elif normalization_method == 'arpb':
        assert type(per_base_depth) is list
        normalized_coverage = statistics.mean(per_base_depth)
    return normalized_coverage


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description=__doc__,
                                     formatter_class=argparse.
                                     RawDescriptionHelpFormatter)
    parser.add_argument('gff3', metavar='GFF3',
                        default=None,
                        nargs='?',
                        help='PROKKA GFF3 file with annotations for the same'
                             + ' metagenome as the FASTA or FASTR file')
    parser.add_argument('output', metavar='out_prefix',
                        default=None,
                        nargs='?',
                        help='name of GFF3 file to write')
    parser.add_argument('--bam', metavar='BAM',
                        default=None,
                        nargs='?',
                        help='BAM file containing mapping data')
    parser.add_argument('-d', '--database', metavar='DB',
                        default='ko',
                        help='extract hits from database "DB" [default: ko]')
    parser.add_argument('--fasta', metavar='FASTA',
                        default=None,
                        nargs='?',
                        help='IDBA generated FASTA file with read count data')
    parser.add_argument('--fastr', metavar='FASTR',
                        default=None,
                        nargs='?',
                        help='FASTR file containing read depth data for a ' +
                             'metatranscriptome mapped onto a metagenome')
    parser.add_argument('-l', '--log_file', metavar='LOG',
                        default=None,
                        help='log file to print all messages to')
    parser.add_argument('--normalize', metavar='normalization method',
                        type=normalization_method,
                        default='read_count',
                        help='how to normailze data [default: read_count]')
    parser.add_argument('-v', '--verbosity',
                        action='count',
                        default=0,
                        help='increase output verbosity')
    parser.add_argument('--verify',
                        action='store_true',
                        help='verify input files before use')
    parser.add_argument('--version',
                        action='store_true',
                        help='prints tool version and exits')
    args = parser.parse_args()

    if args.version:
        print(__version__)
    elif args.fastr is None and args.fasta is None and args.bam is None:
        print(__doc__)
    elif args.gff3 is None:
        message = 'Must specify a GFF3 file and either a BAM, FASTA, ' \
                  'or FASTR file'
        output(message, args.verbosity, 0, log_file=args.log_, fatal=True)
    else:
        if args.verify:

            # Verify BAM file
            if args.bam:
                output('Verifying {0}'.format(args.bam), args.verbosity, 1,
                       log_file=args.log_file)
                with open(args.bam, 'rU') as in_handle:
                    validBam = binary_verifier(in_handle)
                bamValidity = 'valid' if validBam else 'invalid'
                output('{0} is {1}'.format(args.bam, bamValidity),
                       args.verbosity, 1, log_file=args.log_file,
                       fatal=not validBam)

            # Verify FASTA file
            if args.fasta:
                output('Verifying {0}'.format(args.fasta), args.verbosity, 1,
                       log_file=args.log_file)
                with open(args.fasta, 'rU') as in_handle:
                    validFasta = fasta_verifier(in_handle)
                fastaValidity = 'valid' if validFasta else 'invalid'
                output('{0} is {1}'.format(args.fasta, fastaValidity),
                       args.verbosity, 1, log_file=args.log_file,
                       fatal=not validFasta)

            # Verify FASTR file
            if args.fastr:
                output('Verifying {0}'.format(args.fastr), args.verbosity,
                       1, log_file=args.log_file)
                with open(args.fasta, 'rU') as in_handle:
                    validFastr = fastr_verifier(in_handle)
                fastrValidity = 'valid' if validFastr else 'invalid'
                output('{0} is {1}'.format(args.fastr, fastrValidity),
                       args.verbosity, 1, log_file=args.log_file,
                       fatal=not validFastr)

            # Verify GFF3 file
            output('Verifying {0}'.format(args.gff3), args.verbosity, 1,
                   log_file=args.log_file)
            with open(args.gff3, 'rU') as in_handle:
                validGff3 = gff3_verifier(in_handle)
            gff3Validity = 'valid' if validGff3 else 'invalid'
            output('{0} is {1}'.format(args.gff3, gff3Validity),
                   args.verbosity, 1, log_file=args.log_file,
                   fatal=not validGff3)

        # Main portion of program
        # Write header to output
        with open(args.output, 'w') as out_handle:
            out_handle.write('Database_ID\tCoverage_{0}\n'.format(
                                                          args.normalize))
        if args.bam:
            message = 'Computing gene abundances from BAM file {0}'.format(
                                                                   args.bam)
            output(message, args.verbosity, 1, log_file=args.log_file)
            compute_gene_abundance_from_fasta(args.bam,
                                              args.gff3,
                                              args.database,
                                              args.normalize,
                                              args.output)
        elif args.fasta:
            message = 'Computing gene abundances from IDBA-UD generated FASTA' \
                      ' file {0}'.format(args.fasta)
            output(message, args.verbosity, 1, log_file=args.log_file)
            compute_gene_abundance_from_fasta(args.fasta,
                                              args.gff3,
                                              args.database,
                                              args.normalize,
                                              args.output)
        elif args.fastr:
            message = 'Computing gene abundances from FASTR file {0}'.format(
                                                                     args.fastr)
            output(message, args.verbosity, 1, log_file=args.log_file)
            compute_gene_abundance_from_fastr(args.fastr,
                                              args.gff3,
                                              args.database,
                                              args.normalize,
                                              args.output)
        sys.exit(0)



#! /usr/bin/env python

from setuptools import setup

setup(name='metameta',
      version='0.0.0.43',
      description='Toolkit for analyzing '
                  + 'meta-transcriptome/metagenome mapping data',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
          'Natural Language :: English',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 3.4',
          'Topic :: Scientific/Engineering :: Bio-Informatics'
      ],
      keywords='bioinformatics metadata metagenome metatranscriptome'
               + 'short reads mapping alignment',
      url='https://github.com/Brazelton-Lab/metameta/',
      download_url='https://github.com/Brazelton-Lab/metameta/tarball/'
                   + '0.0.0.43',
      author='Alex Hyer',
      author_email='theonehyer@gmail.com',
      license='GPL',
      packages=['metameta', 'metameta.bin', 'metameta.metameta_utils',
                'metameta.data'],
      package_data={
          'data': ['*.txt'],
      },
      zip_safe=False,
      install_requires=[
          'bio_utils',
          'pysam',
          'screed',
          'statistics'
      ],
      entry_points={
          'console_scripts': ['metameta = metameta.bin.__main__:main']
      },
      )
